CON

'//////////////////////////////////////////////////////////////////////////////
' PROP CHIP / SYSTEM DEFINES //////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' These settings are for 5 MHZ XTALS
  _clkmode = xtal1 + pll16x         ' enable external clock and pll times 16
  _xinfreq = 5_000_000              ' set frequency to 5 MHZ

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000
  
  _stack   = 128                    ' accomodate display memory and stack

'//////////////////////////////////////////////////////////////////////////////
' C3 PIN MAP DEFINES //////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  ' status/debug pin also controls VGA buffer enable
  STATUS_LED_BUS_MUX     = 15 ' pin that controls VGA enable as well as a status/debug LED

  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

  ' PS/2 port pin defines
  PS2_DATA               = 26 ' PS/2 port clock pin
  PS2_CLK                = 27 ' PS/2 data pin               

  ' Port A pin defines
  PORTA0                 = 0
  PORTA1                 = 1
  PORTA2                 = 2
  PORTA3                 = 3
  PORTA4                 = 4
  PORTA5                 = 5
  PORTA6                 = 6
  PORTA7                 = 7

  ' Servo port 3-pin header output pins (shared with Port A P4..P7)
  ' note 1: each I/O pins is buffered thru a 50 ohm resistor to minimize long transmission line effects from cables
  ' note 2: each "servo" port has a 3 pin header [signal, power, ground], each power is feed from a jumper adjacent to the
  ' servo port and has a 22uF cap across it as a current source, thus the "servo" ports aside from using the servo 3-pin cable
  ' pin outs, CAN drive servos!!! Care must be taken not to sink too much current, if you find that the board resets under harsh
  ' conditions while driving multiple simulataneous servos then you can add a bulk 2200uF cap for energy storage. There are two holes
  ' pre-drilled by the 2.1mm power entry port, simply solder a 2200uF cap under or above the board, just make sure to get the polarity
  ' correct: minus to the left (next to left edge of board), plus to the right (next to switch).
  SERVO_P4               = 4 
  SERVO_P5               = 5
  SERVO_P6               = 6
  SERVO_P7               = 7


  ' Port B pin defines (muxed with VGA port)
  PORTB0                 = 16  
  PORTB1                 = 17
  PORTB2                 = 18
  PORTB3                 = 19
  PORTB4                 = 20
  PORTB5                 = 21
  PORTB6                 = 22
  PORTB7                 = 23

  ' VGA port pin defines
  VGA_HSYNC              = 16
  VGA_VSYNC              = 17
  VGA_B0                 = 18
  VGA_B1                 = 19
  VGA_G0                 = 20
  VGA_G1                 = 21
  VGA_R0                 = 22
  VGA_R1                 = 23

  ' NTSC/PAL composite video pin defines
  COMPVID0               = 12
  COMPVID1               = 13
  COMPVID2               = 14

  ' Audio output pin define (PWM)
  AUDIO_MONO             = 24

  ' I2C pin defines (standard Prop pins)
  I2C_SCL                = 28
  I2C_SDA                = 29
 
  ' TX/RX pin defines (standard Prop pins)
  UART_TX                = 31
  UART_RX                = 30

  ' ASCII codes for ease of parser development
  ASCII_A      = 65
  ASCII_B      = 66
  ASCII_C      = 67
  ASCII_D      = 68
  ASCII_E      = 69
  ASCII_F      = 70
  ASCII_G      = 71
  ASCII_H      = 72
  ASCII_O      = 79  
  ASCII_P      = 80
  ASCII_Z      = 90
  ASCII_0      = 48
  ASCII_9      = 57
  ASCII_LEFT   = $C0
  ASCII_RIGHT  = $C1
  ASCII_UP     = $C2
  ASCII_DOWN   = $C3 
  ASCII_BS     = $C8 ' backspace
  ASCII_DEL    = $C9 ' delete
  ASCII_LF     = $0A ' line feed 
  ASCII_CR     = $0D ' carriage return
  ASCII_ESC    = $CB ' escape
  ASCII_HEX    = $24 ' $ for hex
  ASCII_BIN    = $25 ' % for binary
  ASCII_LB     = $5B ' [ 
  ASCII_SEMI   = $3A ' ; 
  ASCII_EQUALS = $3D ' = 
  ASCII_PERIOD = $2E ' .
  ASCII_COMMA  = $2C ' ,
  ASCII_SHARP  = $23 ' #
  ASCII_NULL   = $00 ' null character
  ASCII_SPACE  = $20 ' space
  ASCII_TAB    = $09 ' horizontal tab

  ' null pointer, null character
  NULL         = 0  

'//////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY DEFINES ////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

' defines for the proper FLASH installed on C3 4MB, 8MB, 16MB
' 4Mb (512K Bytes)
AT26F004        = 4  ' numeric codes for flash version
AT25DF041A      = 4

' 8Mb (1M Bytes)
AT26DF081A      = 8
AT25DF081A      = 8

' 16 Mb (2M Bytes)
AT25DF161       = 16

' set type of FLASH ROM currently installed in C3 (8MBit) for conditional compilation of code -- not supported in SPIN, so manually done below :( 
FLASH_ROM             = AT26DF081A 
NUM_FLASH_ROM_SECTORS = 19 ' 11 for 4MB, 19 for 8MB, must be set manually due to lack of conditional compilation


' Atmel 2F004/8xxx series basic commands
WRITE_ENABLE     = $06
WRITE_DISABLE    = $04
READ_STATUS      = $05
WRITE_STATUS     = $01
READ_DATA        = $03

SEQ_PROGRAM      = $AF  
BYTE_PROGRAM     = $02 

BLOCK_ERASE4     = $20
BLOCK_ERASE32    = $52
BLOCK_ERASE64    = $D8

SECTOR_ERASE     = $20    

CHIP_ERASE       = $60 
JEDEC_ID         = $9F

PROTECT_SECTOR   = $36 
UNPROTECT_SECTOR = $39 

' custom commands we make up, not part of chip spec, just used here, start at code 0xF0
UNPROTECT_CHIP   = $F0

  ' SRAM function helper constants
  SPI_MODE_BYTE       = %00_00000_1
  SPI_MODE_PAGE       = %10_00000_1
  SPI_MODE_SEQUENTIAL = %01_00000_1


  DO = 10
  CLK =11
  DI = 9
  CS = 3

' SPI (serial peripheral interface) interface pins
  spiDO  = 10
  spiCLK = 11
  spiDI  = 9

   
'//////////////////////////////////////////////////////////////////////////////
' VAR SECTION /////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

  
VAR

 
DAT

'  conditional compilation for FLASH memory variants not supported, so just comment out either declaration

' #if (FLASH_ROM == AT26F004)

'// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
'// 11 sectors [64K, 64K, 64K, 64K, 64K, 64K, 64K, 32K, 8K, 8k, 16K ]
'//            |<------- (7) 64K sectors ------->|  1      2     1 

{
    ' 11 sectors in 4MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 32K sector 7
                     long $78000   ' 8K sector 8  
                     long $7A000   ' 8K sector 9
                     long $7D000   ' 8K sector 10
}

' #if (FLASH_ROM == AT26DF081A)

' we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
' 19 sectors [64K, 64K, 64K, ..., 64K, 64K, 64K, 64K, 16K, 8K, 8k, 32K ]
'            |<----------- (15) 64K sectors -------->| 1     2     1 

    ' 19 sectors in 8MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 64K sector 7 
                     long $80000   ' 64K sector 8
                     long $90000   ' 64K sector 9
                     long $A0000   ' 64K sector 10
                     long $B0000   ' 64K sector 11
                     long $C0000   ' 64K sector 12
                     long $D0000   ' 64K sector 13
                     long $E0000   ' 64K sector 14
                          
                     long $F0000   ' 16K sector 15
                     long $F4000   '  8K sector 16
                     long $F6000   '  8K sector 17
                          
                     long $F8000    ' 32K sector 18


PUB AD_Init

{{
This function initializes the MCP3202 A/D converter. Currently, does nothing
more than asserts the CS then de-asserts it.

PARMS:  None.

RETURNS: Nothing. 

}}

  ' select and de-select the A/D device on the SPI bus just to wake it up and
  ' make sure we know the state of the A/D  
  SPI_Select_Channel( 0 )
  SPI_Select_Channel( SPI_CHAN_AD )
  SPI_Select_Channel( 0 )

' end AD_Init
PUB AD_Read( channel ) | _data

{{
This function reads from the MCP3202 the sent AD channel (0,1) in single ended mode and returns a
12-bit integer [0..4095] representing the value.

PARMS:  channel - [0,1] which channel to read. 

RETURNS: returns value of A/D converter [0..4095] 

}}

  ' read channel 0, SPI transaction is 17 bits total, control bits indicate single ended mode, channel (0,1)
  SPI_Select_Channel( SPI_CHAN_AD )
  _data := SPI_Write_Read( 17, %0001_10_1_0000000000000 | channel << 14, $FFF )
  SPI_Select_Channel( 0 )

  ' return A/D value
  return ( _data )
   
' end AD_Read

PUB SRAM_Init ( mode ) | _spi_word
{{
This function initializes both banks of the Microchip 23K256 SRAMs
serial SPI SRAMs'. The initialization simply sets the STATUS register
of each SRAM into "byte", "page" or "sequential" read/write mode to allow more
efficient access of the SRAMs. The SRAM API currently relies on "sequential"
mode, so if yoo want to use them, this function should be called with SPI_MODE_SEQUENTIAL

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  mode - %00_00000_1 = SPI_MODE_BYTE
               %10_00000_1 = SPI_MODE_PAGE
               %01_00000_1 = SPI_MODE_SEQUENTIAL

RETURNS: Nothing. 

}}

  ' first set SRAM bank 0 to sequential access mode (see page 6 of data sheet)
  SPI_Select_Channel( SPI_CHAN_SRAM0 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' now select SRAM bank 1
  SPI_Select_Channel( SPI_CHAN_SRAM1 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' and finally de-select all SPI devices
  SPI_Select_Channel( 0 )

' end SRAM_Init
PUB SRAM_Write( bank, addr, num_bytes, src_buffer_ptr) | _index, _data, _spi_word
{{
This function writes a number of bytes to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        src_buffer_ptr - pointer to the buffer of bytes to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to write?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write sequential byte:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' and now send command to SRAM for sequential writing
    SPI_Write_Read( 24, _spi_word, $FF )
    
    ' write the bytes into the SRAM
    repeat _index from 0 to num_bytes-1

      ' get byte to write
      _data := byte[ src_buffer_ptr][ _index ]
       
      ' write the byte
      SPI_Write_Read( 8, _data, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Write
PUB SRAM_Fill( bank, addr, num_bytes, value) | _index, _data, _spi_word
{{
This function writes a constant value to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        value          - byte value to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to fill?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write byte sequential:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' setup for sequential byte write, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to receive bytes (as many as we want to send)
    repeat _index from 0 to num_bytes-1
      SPI_Write_Read( 8, value, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Fill

PUB SRAM_Read(bank, addr, num_bytes, dest_buffer_ptr) | _index, _data, _spi_word
{{

This function reads a number of bytes from either bank of the SRAM into a buffer.

The 23K256 supports a number of commands, we have only implemented a couple here
in these API functions (single byte read/write, multiple byte read/write).
However, there are many others you can implement, find out more by reviewing
the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  bank            - [0,1] the SRAM bank to read from, there are two banks, each holds 32K bytes
        addr            - address in SRAM to read from [0..32767]
        num_bytes       - number of bytes to read from SRAM into destination buffer
        dest_buffer_ptr - pointer to the buffer of bytes to read SRAM into

RETURNS: number of bytes read 

}}
 ' test for anything to read?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( read byte sequential:8 | address:16 )
    _spi_word := (%00000011 << 16) + (addr)

    ' setup for sequential byte read, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to be read from
    repeat _index from 0 to num_bytes-1

      ' read next byte from SRAM into receiving buffer
      byte [dest_buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )
    
    ' return bytes read
    return( num_bytes )
  else
    ' catch error
    return 0 

' end SRAM_Read

PUB SRAM64K_MREAD( addr ) | _spi_word, _data
{{
This function reads a single byte from either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Read function
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
    
RETURNS: The 8-bit data from the SRAM. 

}}


    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( read byte sequential:8 | address:16 | dummy:8 )
    _spi_word := (%00000011 << 24) + (addr << 8) + 0 

    ' read the byte of data 
    _data := SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return data
    return ( _data )
    
' end SRAM64K_READ

PUB SRAM64K_MWRITE( addr, data ) | _spi_word
{{
This function writes a single byte to either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Write|Fill functions
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
        data - the byte to write.    

RETURNS: Nothing. 

}}


    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( write sequential byte:8 | address:16 | data: 8)
    _spi_word := (%00000010 << 24) + ( (addr & $7FFF) << 8) + data

    ' write the byte of data 
    SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

' end SRAM64K_WRITE


PUB Flash_Read_Status | _spi_data
{{
Reads the FLASH memory status register

PARMS: None.
RETURNS: status register value
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, READ_STATUS, $FF )

  ' now read data returned
  _spi_data := SPI_Write_Read( 8, $FF, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return results
  return ( _spi_data )   

' // end Flash_Read_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write_Status ( status ) | _spi_data
{{
Writes the FLASH memory status register

PARMS:

status - value to update the status register with

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_STATUS, $FF )
  _spi_data := SPI_Write_Read( 8, status, $FF )  
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return success for now
  return ( 1 )   

' // end Flash_Write_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Open | _index, _spi_data, _flash_start_addr
{{

This functions "opens" the FLASH memory simply by unprotecting all sectors starting at address 0

PARMS: None.
RETURNS: None.

}}


  ' use global unprotect to unprotect entire chip at once, write "0" to bits 2,3,4,5
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
  SPI_Select_Channel( 0 )

  Flash_Write_Status ( Flash_Read_Status & %11_0000_11 )
  Delay_MS( 10 )   
  ' // return success always
  return( 1 )

{
  

  ' this code is used to iterate thru all sectors manually and have more control
  ' iterate thru all sectors and unprotect them
  repeat _index from 0 to NUM_FLASH_ROM_SECTORS-1

    '// STEP 1: enable writing to chip  ///////////////////////////////////////
    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    Delay_MS( 10 )

    '// unprotect sector  ///////////////////////////////////

    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, UNPROTECT_SECTOR, $FF )

    '// get sector address
    _flash_start_addr := sector_addresses[ _index ]

    '// write sector address, there are only a handful of sectors since each is 4-64K
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    ' delay 10
    Delay_MS( 10 ) 

    ' // end for index

  ' // return success always
  return( 1 )
 }

' // end Flash_Open

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Close | _index, _spi_data, _flash_start_addr      
{{
This function "closes" the FLASH by protecting all of the sectors on the flash chip.

PARMS: None.
RETURNS: None.

}}

  '// protecting the sectors seems to have a side effect, we are not considering???
  '// we need to read the data sheet more, for now, just return, and don't protect 
  '// all the sectors, this is overkill anyway for now ... 
  return(1)


' // end Flash_Close

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Erase_Chip( wait ) | _spi_data
{{
This function erases the entire chip. The erase command takes a considerable amount
of time, thus this function issues the command and either returns immediately or waits.
To determine if the erase is complete the function polls the status register bit 0
which is the RDY/BSY bit, 0=device is idle, 1=device is busy

PARMS: wait - boolean, 1=wait for erasure to complete, 0-return immediately.

RETURNS: None.

}}
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, CHIP_ERASE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' test if user requested wait for completion?
  if (wait)
    repeat
      ' erase operation still ongoing?
      if ((Flash_Read_Status & $01)==0)
        return (1)
     
  '// return success for now
  return(1)

' /////////////////////////////////////////////////////////////////////////////


PUB Flash_Erase( block ) | _index, _spi_data, _flash_start_addr
{{
This function erases a block of FLASH memory.

PARMS:

block - block address to erase.

RETURNS: None.

}}
  '// STEP 1: enable writing to chip again /////////////////////////////////////////////////////
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
  Delay_MS( 100 )
   
  '// STEP 2: erase 4K block of chip
  _flash_start_addr := FLASH_BLOCK_TO_ADDR ( block )
   
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, BLOCK_ERASE4, $FF )
   
  '// write block address, there are only a handful of blocks since each is 4-64K
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// erasure takes a long time... 
  Delay_MS( 500 )
   
  '// return success for now
  return(1)
   
' // end Flash_Erase

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function writes num_bytes bytes to the FLASH memory at flash_start_addr from the sent buffer_ptr.

It uses the byte/page write function. Although, there is a "sequential" byte write function available
on the AT26DF081A the new rev of the Atel chips (AT25xxxx series) coming out in 2011 have replaced the sequential write with
another function :(, thus we simply have to check for page boundaries and re-issue another page write
every 256 bytes.

PARMS:

flash_start_addr - starting address in the FLASH memory to write the data at.
buffer_ptr       - local byte pointer to Prop memory holding data to write to FLASH memory.
num_bytes        - number of bytes to write to FLASH from source buffer_ptr. 

}}

  '// STEP 1: enable writing to the chip again
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// STEP 2: write buffer to address in flash //////////////////
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
   
  '// initiate byte/page program command
  _spi_data := SPI_Write_Read( 8, BYTE_PROGRAM, $FF )
   
  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )

  '// write first byte of 
  _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ 0 ], $FF )

  _index := 1   

  '// write remaining bytes
  repeat until (_index => num_bytes )

    ' now test if we are on a page boundary, if so finish page, and start new page
    if ( ((flash_start_addr + _index) // 256) == 0)
      '// set CS to SPI select channel 0 (null)
      SPI_Select_Channel( 0 )

      ' replace this with polling...
      Delay_MS(10) 

      '// set CS to SPI select channel 3 (FLASH)
      SPI_Select_Channel( SPI_CHAN_FLASH )
      _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
      SPI_Select_Channel( 0 )

      SPI_Select_Channel( SPI_CHAN_FLASH )
               
      '// re-issue byte/page program command
      _spi_data := SPI_Write_Read( 8,BYTE_PROGRAM,$FF )

      '// write starting address byte by byte, MSB first...
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 16, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 8, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 0, $FF )

      ' end if 

    '// write the next byte of data
    _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ _index ], $FF )

    _index++
    
    ' // end for index    

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' write final bytes left in internal buffer, replace this with polling...
  Delay_MS(10) 
     
  '// disable SPI interface
   
  '// return success
  return(1)
   
' // end Flash_Write

' ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Flash_Read( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function reads num_bytes from the FLASH memory starting at memory location flash_start_addr and stores the data in
buffer_ptr.

PARMS:

flash_start_addr - starting address in FLASH memory to read data from.
buffer_ptr       - local Prop byte pointer to store the read data at.
num_bytes        - the number of bytes to read from the FLASH memory. 

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )

  '// initiate read command
  _spi_data := SPI_Write_Read( 8, READ_DATA, $FF )

  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )
  
  '// now read data 
  repeat _index from  0 to num_bytes-1
    byte [buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' // end for index

  '// disable SPI interface

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// return success 
  return(1)

' // end Flash_Read

' /////////////////////////////////////////////////////////////////////////////

'//////////////////////////////////////////////////////////////////////////////
' FLASH MACROS ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB FLASH_ADDR_TO_BLOCK( n )
  n := (n >> 12)

' /////////////////////////////////////////////////////////////////////////////
  
PUB FLASH_BLOCK_TO_ADDR( n )
  n := (n << 12)

' /////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////
' SPI ROUTINES///////// ///////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Init
' DESCRIPTION: this function initializes the hardware interface to the SPI bus. on the HYDRA SD MAX card
' (as with most SPI buses) there are only 4 signals; clk, data in, data out, and chip select

' INPUTS: mode - c ontrols the phase of the clock and the logic levels
' mode 0 - default support 
' mode 1 - not supported
' mode 2 - not supported
' mode 3 - not supported
'
' OUTPUTS: none 
' -----------------------------------------------------------------------------

  ' set directions of SPI interface, be careful not to clock the SPI interface accidentally
  ' also, data in and data out are referenced from TARGET, so DI means the input to the SPI device which would be
  ' an OUTPUT from the propeller
   OUTA [ spiDI ]  := 0 ' set output LOW
   DIRA [ spiDI ]  := 1 ' set "data in" to output

   DIRA [ spiDO  ] := 0 ' set "data out" to input

   OUTA [ spiCLK ] := 0 ' set output LOW
   DIRA [ spiCLK ] := 1 ' set "clock" to output 

   ' set SPI mux counter IOs up
   DIRA[SPI_SEL_CLK] := 1 ' set to output
   OUTA[SPI_SEL_CLK] := 0
  
   DIRA[SPI_SEL_CLR] := 1 ' set to output
   OUTA[SPI_SEL_CLR] := 0 ' CLR counter
   OUTA[SPI_SEL_CLR] := 1 ' allow counting

'   OUTA [ spiCS ]  := 1 ' set output HIGH (de-select the device)
'   DIRA [ spiCS ]  := 1 ' set "chip select" to output

' end SPI_Init

'//////////////////////////////////////////////////////////////////////////////

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel



PUB SPI_Send_Recv_Byte(spi_data8) | spi_result8, index

' DESCRIPTION: this functions sends spi_data8 and receives spi_result8, remember all spi operations are circular, so
' when a byte is sent a byte is received at the same time, you can ignore the sent or received byte as you
' wish. For example, to do a read just send the dummy value $FF
' NOTE: assumes caller is controlling the chip select line
'
' INPUTS: spi_data8 - 8 bit data that is going to be sent
' OUTPUTS: returns the 8-bit data received at the same time
' -----------------------------------------------------------------------------

  ' reset result
  spi_result8 := 0

  ' shift 8-bits of data out while shifting in 8-bits of data
  ' data is sent and receiving MSB to LSB, so bit 7,6,5,4,3,2,1,0
  ' assume SPI interface is in mode 0/1, later make more flexible 
  repeat index from 0 to 7
    ' place next bit on data in pin to device
    OUTA [ spiDI ] := ((spi_data8 & $80) >> 7)
'   vga.dec((spi_data8 & $80) >> 7)
    
    spi_data8 <<= 1 ' shift data to left and prepare next bit for transmission

  ' pulse the clock line
    OUTA [ spiCLK ] := 1

      ' data is ready now on output line from device
    spi_result8 <<= 1            ' shift result to left
    spi_result8 |= INA [ spiDO ] ' read data bit and OR into result
                                    
    ' finish clock pulse
    OUTA [ spiCLK ] := 0

  ' return result
  return (spi_result8)

' end SPI_Send_Receive_Byte

'//////////////////////////////////////////////////////////////////////////////

PUB SPI_Read_Byte

' DESCRIPTION: reads a single byte from the SPI interface, simply a dummy call to SPI_Send_Recv_Byte
'
' INPUTS: none. 
'
' OUTPUTS: returns the byte read from the SPI interface
' -----------------------------------------------------------------------------

  return ( SPI_Send_Recv_Byte($FF) )

' end SPI_Read_Byte

'//////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Byte(spi_data8)

' DESCRIPTION: sends a single byte to the SPI interface, returns the read value
'
' INPUTS: spi_data8 - 8-bit data to send 
'
' OUTPUTS: returns the 8-bit data from the SPI interface
' -----------------------------------------------------------------------------

' return the data from the cylic buffer regardless...
return( SPI_Send_Recv_Byte(spi_data8) )

' end SPI_Write_Byte

' /////////////////////////////////////////////////////////////////////////////
' UTILITY ROUTINES ////////////////////////////////////////////////////////////
' /////////////////////////////////////////////////////////////////////////////

PUB StrCpy( dest_str_ptr, source_str_ptr ) | _index
{{
mimics C strcpy, copies source -> destination including NULL
}}

' test if there is storage
if (dest_str_ptr == NULL)
  return (NULL)

_index := 0

repeat while (byte [ source_str_ptr ][_index] <> NULL)
  ' copy next byte
   byte [ dest_str_ptr ][_index] := byte [ source_str_ptr ][_index]    
   _index++

' null terminate
byte [ dest_str_ptr ][_index] := NULL

' return number of bytes copied
return ( _index )

' // end StrCpy

' /////////////////////////////////////////////////////////////////////////////

PUB StrUpper(string_ptr)
{{
upcases a string
}}

  if (string_ptr <> NULL)
    repeat while (byte[ string_ptr ] <> NULL)
      byte[ string_ptr ] :=  ToUpper( byte[ string_ptr ] )
      string_ptr++  

  ' return string
  return (string_ptr)

' // end StrUpper

' /////////////////////////////////////////////////////////////////////////////


PUB IsInSet(ch, set_string)
{{
tests if sent character is in string
}}

repeat while (byte[set_string][0] <> NULL)
  if (ch == byte[set_string++][0])
    'serial.txstring(string ("found!") )
    return( ch )

'serial.txstring(string ("not found!") )
' not found
return( -1 )    

' // end IsInSet

' /////////////////////////////////////////////////////////////////////////////

PUB IsSpace(ch)
{{
tests if sent character is white space, cr, lf, space, tab
}}

if ( (ch == ASCII_SPACE) or (ch == ASCII_LF) or (ch == ASCII_CR) or (ch == ASCII_TAB))
  return (ch)
else
  return(-1)  

' // end IsSpace

' /////////////////////////////////////////////////////////////////////////////

PUB IsNull(ch)
{{
tests if sent character is null
}}

if ( (ch == NULL))
  return (1)
else
  return(-1)  

' // end IsNull

' /////////////////////////////////////////////////////////////////////////////

PUB IsDigit(ch)
{{
tests if sent character is a number 0..9, returns integer 0..9
}}

if (ch => ASCII_0 and ch =< ASCII_9)
  return (ch-ASCII_0)
else
  return(-1)  

' // end IsDigit

' /////////////////////////////////////////////////////////////////////////////

PUB IsAlpha(ch)
{{
tests if sent character is a number a...zA...Z
}}

ch := ToUpper(ch)

if ( (ch => ASCII_A) and (ch =< ASCII_Z))
  return (ch)
else
  return(-1)  

' end IsAlpha

' /////////////////////////////////////////////////////////////////////////////

PUB IsPunc(ch)
{{
tests if sent character is a punctuation symbol !@#$%^&*()--+={}[]|\;:'",<.>/?
}}

ch := ToUpper(ch)

if (  ((ch => 33) and (ch =< 47)) or  ((ch => 58) and (ch =< 64)) or ((ch => 91) and (ch =< 96)) or ((ch =>123) and (ch =< 126))  ) 
  return (ch)
else
  return(-1)  

' // end IsPunc
                                                                                
' /////////////////////////////////////////////////////////////////////////////

PUB HexToDec(n)
{{
converts hex digit to decimal
}} 
  if ( (n => "0") and (n =< "9") )
    return (n - ASCII_0)
  elseif ( (n => "A") and (n =< "F") )
    return (n - "A" + 10)
  else
    return ( 0 )  

' // end HexToDec

' /////////////////////////////////////////////////////////////////////////////

PUB itoa(value, string_ptr) | i
{{
 converts value into a string and stores in string_ptr with NULL termination
 }}

  if value < 0
    -value
    byte [ string_ptr++ ]  := "-"

  i := 1_000_000_000

  repeat 10
    if value => i
      byte [ string_ptr++ ]  := value / i + "0"
      value //= i
      result~~
    elseif result or i == 1
      byte [ string_ptr++ ]  := "0"
    i /= 10

  ' null terminate
  byte [ string_ptr ] := NULL    

' // end itoa

' /////////////////////////////////////////////////////////////////////////////

PUB atoi2(string_ptr, length) | _index, _sum, _ch, _sign
{{
this function tries to convert the string to a number, supports binary %, hex $, decimal (default)
string_ptr can be to a null terminated string, or the length can be overridden by sending the length shorter
than the null terminator, ignores ws
Eg. %001010110, $EF, 25
}}

' initialize vars
_index := 0
_sum   := 0
_sign  := 1
 
' consume white space
repeat while (IsSpace( byte[ string_ptr ][_index] ) <> -1)
  _index++
 
' is there a +/- sign?
if (byte [string_ptr][_index] == "+")
  ' consume it
  _index++
elseif (byte [string_ptr][_index] == "-")
  ' consume it
  _index++
  _sign := -1    
     
' try to determine number base
if (byte [string_ptr][_index] == ASCII_HEX)
  _index++
  repeat while ( ( IsDigit(_ch := byte [string_ptr][_index]) <> -1) or ( IsAlpha(_ch := byte [string_ptr][_index]) <> -1)  )
    _index++
    _sum := (_sum << 4) + HexToDec( ToUpper(_ch) )
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if hex number    
elseif (byte [string_ptr][_index] == ASCII_BIN)
  repeat while ( IsDigit(_ch := byte [string_ptr][++_index]) <> -1)
    _sum := (_sum << 1) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
' // end if binary number  
else
  ' must be in default base 10, assume that
  repeat while ( IsDigit(_ch := byte [string_ptr][_index++]) <> -1)
    _sum := (_sum * 10) + (_ch - ASCII_0)
    if (_index => length)
      return (_sum*_sign)
 
  return(_sum*_sign)
 
' else, have no idea of number format!
return(0)
 
' // end atoi2

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US





PUB atoi(string_ptr) | sign, factor, number
' DESCRIPTION: converts the sent string to an integer and returns it, assumes the string is
' an integer in ASCII format, only flexibility function has is it will strip leading spaces
'
' INPUTS:
' OUTPUTS:
' -----------------------------------------------------------------------------

  sign   := 1
  number := 0

  ' if string is NULL return 0
  if (string_ptr == 0)
    return (0) 

  ' strip leading whitespace (spaces only)
  repeat while (byte[string_ptr] == $20)
    string_ptr++
    
  ' test for negative sign "-"
  if (byte[string_ptr]==$2D)
    sign := -1
    string_ptr++

  ' now convert the number
  repeat while ( (byte[string_ptr] => $30) and (byte[string_ptr] =< $39) ) 
    number := 10*number + (byte[string_ptr++] - $30)

  return ( number*sign )

' end atoi




PUB ToUpper(char)

' DESCRIPTION: this function upcases the sent character, if its not lowercase character then it passes it thru
' INPUTS: char - the character to upcase
' OUTPUTS: the upcased character
' -----------------------------------------------------------------------------
 
  if ( (char => 97) and (char =< 122))
    char -= 32
  return (char)

' /////////////////////////////////////////////////////////////////////////////

PUB Strncomp(string_ptr1, string_ptr2, length) | index, char

' DESCRIPTION: this function compares the two sent strings up to the nth character
' the compare is case insensitive
' INPUTS: string_ptr1 - 1st NULL terminated string to compare
'         string_ptr2 - 2ne NULL terminated string to compare 
' OUTPUTS: returns 1 if strings equal, 0 if not (case insensitive)
' -----------------------------------------------------------------------------
  
  repeat index from 0 to length-1
    ' characters are the same?
    if ( ToUpper(byte[string_ptr1][index]) <> ToUpper(byte[string_ptr2][index]) )
      ' strings differ, return failure
      return( 0 )

  ' strings must be equal      
  return( 1 )
' end Strncomp 

' /////////////////////////////////////////////////////////////////////////////



PUB To_ASCII(inchar, replace_char)

' DESCRIPTION: converts input into a ASCII printable character, this macro function is
' is used to simply turn non-printable garbage to printable text, so the tv terminal doesn't
' get sent garbage.
'
' INPUTS: inchar - input character to test.
'         replace_char - if the input char isn't printable then replace with this character
'         (usually a dot or space, so you can see what changed) 
'
' OUTPUTS: returns the converted character
' -----------------------------------------------------------------------------

  ' test the character for printable range
  if (inchar < $20 or inchar > $7E)
    inchar := replace_char
    return(inchar)         

' end To_ASCII

' /////////////////////////////////////////////////////////////////////////////

PUB _Min(a,b)

' DESCRIPTION: returns signed minimum of a,b
' INPUTS: a,b - values to be compared
' OUTPUTS: the minimum
' -----------------------------------------------------------------------------

if (a < b)
  return a
else 
  return b

' /////////////////////////////////////////////////////////////////////////////

PUB _Max(a,b)

' DESCRIPTION: returns signed maximum of a,b
' INPUTS: a,b - values to be compared
' OUTPUTS: the maximum
' -----------------------------------------------------------------------------

if (a > b)
  return a
else 
  return b

'//////////////////////////////////////////////////////////////////////////////

DAT

newline_string          byte      $0D, $0A, $00  ' $0D carriage return, $0A line feed
sram_buffer             byte      1,4,9,16,25,36,49,64,81,100,121,144,169,196,225
{{
 ______________________________________________________________________________________________________________________________
|                                                   TERMS OF USE: MIT License                                                  |                                                            
|______________________________________________________________________________________________________________________________|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |     
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
 ------------------------------------------------------------------------------------------------------------------------------ 
}}